module deimos.event2._opaque_structs;

struct evbuffer;
struct evhttp_request;
struct event_base;

struct evrpc;
struct evrpc_req_generic;
struct evrpc_base;
struct evrpc_request_wrapper;
struct evrpc_status;
struct evrpc_hook_meta;
